/*********************************************************************************
HLS2RGB.java - Converte trs bandas, em nveis de cinza, do espao de cores HLS 
               (Hue, Lightness, Saturation), para o espao de cores RGB (Red, 
               Green, Blue).

Direitos Autorais Reservados (c) 2007 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

**********************************************************************************

Parte integrante do software 'Fuso espectral de imagens de satlite no espao de cores'

Descrio:

    Converte uma as bandas HLS para o espao de cores RGB.

Parmetros de entrada:

    1. banda corresponde ao espectro do Hue;
    2. banda corresponde ao espectro do Lightness;
    3. banda corresponde ao espectro do Saturation.
 
Sada:
 
    1. banda Red;
    2. banda Green;
    3. banda Blue.

Mais informaes, nas pginas 35 do Pitas e 592 do Foley et al.:
    1. Pitas, I. Digital image processing algorithms and applications.
          New York,NY:John Wiley & Sons Inc, 2000. 419p.
    2. Foley, J.D.; A. van Dam; S.K. Feiner; J.F. Hughes. Computer graphics:
          principles and practice, NY:Addison-Wesley, 1990. 1174p.
 
Desenvolvido por:
   
    Marcello Antonio Marabita
    Jos Iguelmar Miranda
 
**********************************************************************************/

// pacotes genricos
import java.io.File;
import java.io.IOException;

// pacotes do AWT
import java.awt.BorderLayout;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.awt.image.Raster;
import java.util.Vector;
import javax.swing.JLabel;
import javax.swing.ImageIcon;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

// leitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;
import javax.swing.JOptionPane;
import javax.swing.JProgressBar;
import javax.swing.JTabbedPane;

public class HLS2RGB extends JPanel {
    
    private JProgressBar barraStatus;
    private boolean sucesso = false;
    
    private void updateBarra(String s, int n) {
        if(n%5 == 0) {
            barraStatus.setString(s);
            barraStatus.setValue(n);
            barraStatus.update(barraStatus.getGraphics());
        }
    }
    
    // CONSTRUTOR DA CLASSE
    
    public HLS2RGB(File fileImgH, File fileImgL, File fileImgS, int tipoMetodo, JProgressBar barra) {
        barraStatus = barra;
        updateBarra("Preparando para Converter, Aguarde!",5);
        
        int numBandas, wH, hH, wL, hL, wS, hS, tipo;
        JLabel img1, img2, img3;
        BufferedImage[] src = new BufferedImage[3];
        BufferedImage[] dest = new BufferedImage[3];
        
        // L as trs bandas em nveis de cinza
        src[0] = abreImagem(fileImgH);    // hue
        src[1] = abreImagem(fileImgL);    // lightness
        src[2] = abreImagem(fileImgS);    // saturation
        
        // Acha as dimenses das imagens...
        wH = src[0].getWidth();
        hH = src[0].getHeight();
        wL = src[1].getWidth();
        hL = src[1].getHeight();
        wS = src[2].getWidth();
        hS = src[2].getHeight();
        
        // ... e checa se elas so iguais
        if ((wH != wL) || (wH != wS) || (hH != hL) || (hH != hS) ||
                (wS != wL) || (hS != hL)) {
            System.out.println("Bandas com dimenses diferentes!");
            JOptionPane.showMessageDialog(this,"No foi possivel realizar esta operao\n" +
                    "Tamanho diferente das bandas!",
                    "Erro: Tamanho das bandas",
                    JOptionPane.ERROR_MESSAGE);
            updateBarra("",0);
            return;
        }
        
        // Certifica que cada banda est em nvel de cinza
        // banda 1...
        numBandas = src[0].getSampleModel().getNumBands();
        if (numBandas > 1) {
            String msgA = "Convertendo para RGB: ";
            String msgB = "\nImagem fornecida para 'Hue' *DEVE* ter 1 banda.";
            String msgC = "#bandas atual = " + numBandas;
            System.out.println(msgA + msgB + "\n" + msgC + ". Abortando...");
            JOptionPane.showMessageDialog(this,msgA+msgB+"\n" + msgC);
            updateBarra("",0);
            return;
        }
        
        // banda 2...
        numBandas = src[1].getSampleModel().getNumBands();
        if (numBandas > 1) {
            String msgA = "Convertendo para RGB: ";
            String msgB = "\nImagem fornecida para 'Lightness' *DEVE* ter 1 banda.";
            String msgC = "#bandas atual = " + numBandas;
            System.out.println(msgA + msgB + "\n" + msgC + ". Abortando...");
            JOptionPane.showMessageDialog(this,msgA+msgB+"\n" + msgC);
            updateBarra("",0);
            return;
        }
        
        // banda 3...
        numBandas = src[2].getSampleModel().getNumBands();
        if (numBandas > 1) {
            String msgA = "Convertendo para RGB: ";
            String msgB = "\nImagem fornecida para 'Saturation' *DEVE* ter 1 banda.";
            String msgC = "#bandas atual = " + numBandas;
            System.out.println(msgA + msgB + "\n" + msgC + ". Abortando...");
            JOptionPane.showMessageDialog(this,msgA+msgB+"\n" + msgC);
            updateBarra("",0);
            return;
        }
        
        updateBarra("Convertendo imagem, Aguarde!",25);
        setName("HLS-RGB: " + fileImgH.getName());
        
        // Transforma espao de cores HLS para RGB
        dest = converteHLS2RGB(src);
        
        updateBarra("Criando a Interface grafica, Aguarde!",70);
        
        JTabbedPane jTabbedPane1 = new JTabbedPane();
        img1 = new JLabel(new ImageIcon(dest[0]));
        img2 = new JLabel(new ImageIcon(dest[1]));
        img3 = new JLabel(new ImageIcon(dest[2]));
        
        jTabbedPane1.addTab("Red",new JScrollPane(img1));
        jTabbedPane1.addTab("Green",new JScrollPane(img2));
        jTabbedPane1.addTab("Blue",new JScrollPane(img3));
        this.setLayout(new BorderLayout());
        this.add(jTabbedPane1);
        updateBarra("Criando a Interface grafica, Aguarde!",75);
        
        // Salva arquivos de sada formato PNG
        File vFile[] = salvaImagem(dest, fileImgH.getName());
        
        Vector vect = new Vector();
        vect.add(jTabbedPane1);
        vect.add(new Manipulador(vFile[0],0,1.00f));
        vect.add(new Manipulador(vFile[1],1,1.00f));
        vect.add(new Manipulador(vFile[2],2,1.00f));
        
        ColorSpaceGui.imagensAbertas.add(vect);
        
        updateBarra("Imagem Convertida: HLS para RGB",100);
        sucesso = true;
    }  // fim Construtor
    
    
    // MTODOS PBLICOS
    
    public BufferedImage[] converteHLS2RGB(BufferedImage[] src) {
        
        int wH, hH, tipo, hue, lightness, saturation, r, g, b;
        double dH, dL, dS, UNDEFINED = 360.0, dR = 0.0, dG = 0.0, dB = 0.0, m1, m2;
        BufferedImage[] out = new BufferedImage[3];
        
        // Como as bandas tm a mesma dimenso, s uma dimenso  suficiente
        wH = src[0].getWidth();
        hH = src[0].getHeight();
        
        // Gerar trs imagens separadas: red, green, blue
        // Posteriormente, elas podem ser combinadas numa imagem falsa cor.
        tipo = BufferedImage.TYPE_BYTE_GRAY;
        BufferedImage biRed   = new BufferedImage(wH, hH, tipo);
        BufferedImage biGreen = new BufferedImage(wH, hH, tipo);
        BufferedImage biBlue  = new BufferedImage(wH, hH, tipo);
        
        Raster srcH = src[0].getRaster();
        Raster srcL = src[1].getRaster();
        Raster srcS = src[2].getRaster();
        
        WritableRaster wrRed   = biRed.getRaster();
        WritableRaster wrGreen = biGreen.getRaster();
        WritableRaster wrBlue  = biBlue.getRaster();
        
        updateBarra("Convertendo imagem, Aguarde!",30);
        System.out.println("Convertendo imagem...");
        
        // Os clculos a seguir normalizam red, green e blue em [0.0, 1.0]
        for (int i = 0; i < hH; i++){
            updateBarra("Convertendo imagem, Aguarde!", (int) (31 + (40*i)/hH));
            
            for (int j = 0; j < wH; j++) {
                hue = srcH.getSample(j, i, 0);
                lightness = srcL.getSample(j, i, 0);
                saturation = srcS.getSample(j, i, 0);
                
                // Normaliza hue, de [0, 255], para [0, 360]. UNDEFINED = 360
                dH = (double)((hue/255.0)*360.0);
                
                // Normaliza lightness e saturation, de [0, 255], para [0.0, 1.0]
                dL = (double)lightness/255.0;
                dS = (double)saturation/255.0;
                
                if (dL <= 0.5)
                    m2 = dL*(1.0 + dS);
                else
                    m2 = dL + dS - dL*dS;
                m1 = 2.0*dL - m2;
                // Caso acromtico, no existe hue, ento red = green = blue = lightness
                if (dS == 0.0) {
                    if (dH == UNDEFINED)
                        dR = dG = dB = dS;
                    else {
                        dR = dG = dB = 0.0;   // Pitas
                    }
                    // Caso cromtico, existe um valor para hue
                } else {
                    dR = value(m1, m2, dH + 120.0);
                    dG = value(m1, m2, dH);
                    dB = value(m1, m2, dH - 120.0);
                }
                
                // Salva imagens...
                // Converte red, green, blue, do intervalo [0.0, 1.0], para [0, 255]
                r = (int)(dR*255);
                g = (int)(dG*255);
                b = (int)(dB*255);
                wrRed.setSample(j, i, 0, r);
                wrGreen.setSample(j, i, 0, g);
                wrBlue.setSample(j, i, 0, b);
            }
        }
        out[0] = biRed;
        out[1] = biGreen;
        out[2] = biBlue;
        
        return out;
    }
    
    // Salva imagens R, G, B no formato .PNG
    public File[] salvaImagem(BufferedImage[] dest, String nomeArquivo) {
        
        String search, prefixo, nome;
        File vFile[] = new File[3];
        
        // Grava imagem.
        search = ".";
        int i = nomeArquivo.indexOf(search);
        prefixo = nomeArquivo.substring(0, i);
        
        try {
            // imagem Red
            updateBarra("Salvando as Imagens, Aguarde!",80);
            nome = "imagens\\" + prefixo + "_RGB_R.png";
            System.out.println("\nGravou arquivo -> " + nome + "\n");
            vFile[0] = new File(nome);
            ImageIO.write(dest[0], "png", vFile[0]);
            
            // imagem Green
            updateBarra("Salvando as Imagens, Aguarde!",85);
            nome = "imagens\\" + prefixo + "_RGB_G.png";
            System.out.println("\nGravou arquivo -> " + nome + "\n");
            vFile[1] = new File(nome);
            ImageIO.write(dest[1], "png", vFile[1]);
            
            // imagem Blue
            updateBarra("Salvando as Imagens, Aguarde!",90);
            nome = "imagens\\" + prefixo + "_RGB_B.png";
            System.out.println("\nGravou arquivo -> " + nome + "\n");
            vFile[2] = new File(nome);
            ImageIO.write(dest[2], "png", vFile[2]);
        } catch (IOException e) {
            System.out.println("Problema gravando arquivo. ");
            System.exit(0);
        }
        
        return vFile;
    }
    
    // MTODOS PRIVADOS DA CLASSE
    
    // Abre arquivo de imagens
    private BufferedImage abreImagem(File aFile) {
        BufferedImage imagem = null;
        
        try {
            imagem = ImageIO.read(aFile);
        } catch(Exception e) {
            System.out.println("abreImagens: Problema lendo '" + aFile + "'.");
        }
        return imagem;
    }
    
    
    
    private double value(double n1, double n2, double hue) {
        double valor;
        
        if(hue > 360.0)
            hue -= 360.0;
        else if(hue < 0.0)
            hue += 360.0;
        if(hue < 60.0)
            valor = n1 + (n2 - n1)*hue/60.0;
        else if(hue < 180.0)
            valor = n2;
        else if(hue < 240.0)
            valor = n1 + (n2 - n1)*(240.0 - hue)/60.0;
        else
            valor = n1;
        
        return valor;
    }
    
    public boolean isSucesso() {
        return sucesso;
    }
}
