/*********************************************************************************
IHS2RGB.java - Converte trs bandas, em nveis de cinza, do espao de cores IHS 
               (Intensity, Hue, Saturation), para o espao de cores RGB (Red, 
               Green, Blue).
 
Direitos Autorais Reservados (c) 2007 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

**********************************************************************************

Parte integrante do software 'Fuso espectral de imagens de satlite no espao de cores'

Descrio:

    Converte uma as bandas IHS para o espao de cores RGB.

Parmetros de entrada:

    1. banda corresponde ao espectro do Intensity;
    2. banda corresponde ao espectro do Hue;
    3. banda corresponde ao espectro do Saturation.
 
Sada:
 
    1. banda Red;
    2. banda Green;
    3. banda Blue.

Mais informaes, nas pginas 35 do Pitas e 592 do Foley et al.:
    1. Pitas, I. Digital image processing algorithms and applications.
          New York,NY:John Wiley & Sons Inc, 2000. 419p.
    2. Foley, J.D.; A. van Dam; S.K. Feiner; J.F. Hughes. Computer graphics:
          principles and practice, NY:Addison-Wesley, 1990. 1174p.
 
Desenvolvido por:
   
    Marcello Antonio Marabita
    Jos Iguelmar Miranda
 
**********************************************************************************/

// pacotes genricos
import java.awt.Color;
import java.io.File;
import java.io.IOException;

// pacotes do AWT
import java.awt.BorderLayout;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.awt.image.Raster;
import java.util.Vector;
import javax.swing.JLabel;
import javax.swing.ImageIcon;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

// leitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;
import javax.swing.JOptionPane;
import javax.swing.JProgressBar;
import javax.swing.JTabbedPane;

public class IHS2RGB extends JPanel {
    
    private JProgressBar barraStatus;
    private boolean sucesso = false;
    
    private void updateBarra(String s, int n) {
        if(n%5 == 0) {
            barraStatus.setString(s);
            barraStatus.setValue(n);
            barraStatus.update(barraStatus.getGraphics());
        }
    }
    
    // CONSTRUTOR DA CLASSE
    
    public IHS2RGB(File imgI, File imgH, File imgS , int tipoMetodo, JProgressBar barra) {
        barraStatus = barra;
        updateBarra("Preparando para Converter, Aguarde!",5);
        
        int numBandas, wI, hI, wH, hH, wS, hS;
        JLabel img1, img2, img3;
        BufferedImage[] src = new BufferedImage[3];
        BufferedImage[] dest = new BufferedImage[3];
        
        // L as trs bandas em nveis de cinza
        src[0] = abreImagem(imgI);    // intensity
        src[1] = abreImagem(imgH);    // hue
        src[2] = abreImagem(imgS);    // saturation
        
        // Acha as dimenses das imagens...
        wI = src[0].getWidth();
        hI = src[0].getHeight();
        wH = src[1].getWidth();
        hH = src[1].getHeight();
        wS = src[2].getWidth();
        hS = src[2].getHeight();
        
        // ... e checa se elas so iguais
        if ((wI != wH) || (wI != wS) || (hI != hH) || (hI != hS) ||
                (wS != wH) || (hS != hH)) {
            System.out.println("Bandas com dimenses diferentes!");
            JOptionPane.showMessageDialog(this,"No foi possivel realizar esta operao\n" +
                    "Tamanho diferente das bandas!",
                    "Erro: Tamanho das bandas",
                    JOptionPane.ERROR_MESSAGE);
            updateBarra("",0);
            return;
        }
        
        // Certifica que cada banda est em nvel de cinza
        // banda 1...
        numBandas = src[0].getSampleModel().getNumBands();
        if (numBandas > 1) {
            String msgA = "Convertendo para RGB: ";
            String msgB = "\nImagem fornecida para 'Intensity' *DEVE* ter 1 banda.";
            String msgC = "#bandas atual = " + numBandas;
            System.out.println(msgA + msgB + "\n" + msgC + ". Abortando...");
            JOptionPane.showMessageDialog(this,msgA+msgB+"\n" + msgC);    
            updateBarra("",0);
            return;
        }
        
        // banda 2...
        numBandas = src[1].getSampleModel().getNumBands();
        if (numBandas > 1) {
            String msgA = "Convertendo para RGB: ";
            String msgB = "\nImagem fornecida para 'Hue' *DEVE* ter 1 banda.";
            String msgC = "#bandas atual = " + numBandas;
            System.out.println(msgA + msgB + "\n" + msgC + ". Abortando...");
            JOptionPane.showMessageDialog(this,msgA+msgB+"\n" + msgC);
            updateBarra("",0);
            return;
        }
        
        // banda 3...
        numBandas = src[2].getSampleModel().getNumBands();
        if (numBandas > 1) {
            String msgA = "Convertendo para RGB: ";
            String msgB = "\nImagem fornecida para 'Saturation' *DEVE* ter 1 banda.";
            String msgC = "#bandas atual = " + numBandas;
            System.out.println(msgA + msgB + "\n" + msgC + ". Abortando...");
            JOptionPane.showMessageDialog(this,msgA+msgB+"\n" + msgC);
            updateBarra("",0);
            return;
        }
        
        updateBarra("Convertendo imagem, Aguarde!",25);
        setName("IHS-RGB: " + imgI.getName());
        
        // Chama funo de converso
        dest = converteIHS2RGB(src);
        
        updateBarra("Criando a Interface grafica, Aguarde!",70);
        
        JTabbedPane jTabbedPane1 = new JTabbedPane();
        img1 = new JLabel(new ImageIcon(dest[0]));
        img2 = new JLabel(new ImageIcon(dest[1]));
        img3 = new JLabel(new ImageIcon(dest[2]));
        
        jTabbedPane1.addTab("Red",new JScrollPane(img1));
        jTabbedPane1.addTab("Green",new JScrollPane(img2));
        jTabbedPane1.addTab("Blue",new JScrollPane(img3));
        this.setLayout(new BorderLayout());
        this.add(jTabbedPane1);
        updateBarra("Criando a Interface grafica, Aguarde!",75);
        
        // Salva arquivos de sada formato PNG
        File vFile[] = salvaImagem(dest, imgI.getName());
        
        Vector vect = new Vector();
        vect.add(jTabbedPane1);
        vect.add(new Manipulador(vFile[0],0,1.00f));
        vect.add(new Manipulador(vFile[1],1,1.00f));
        vect.add(new Manipulador(vFile[2],2,1.00f));
        
        ColorSpaceGui.imagensAbertas.add(vect);
        
        updateBarra("Imagem Convertida: IHS para RGB",100);
        sucesso = true;
    }  // fim Construtor
    
    
    // MTODOS PBLICOS
    
    public BufferedImage[] converteIHS2RGB(BufferedImage[] src) {
        
        int wI, hI, wH, hH, wS, hS, tipo, iI, iH, iS, iR = 0, iG = 0, iB = 0, rgb;
        float intensity, hue, saturation;
        byte bR, bG, bB;
        BufferedImage[] out = new BufferedImage[3];
        
        // Acha as dimenses das imagens...
        wI = src[0].getWidth();
        hI = src[0].getHeight();
  
        // Gerar trs imagens separadas: vermelho, verde, azul
        // Posteriormente, elas podem ser combinadas numa imagem falsa cor.
        tipo = BufferedImage.TYPE_BYTE_GRAY;
        BufferedImage biR = new BufferedImage(wI, hI, tipo);
        BufferedImage biG = new BufferedImage(wI, hI, tipo);
        BufferedImage biB = new BufferedImage(wI, hI, tipo);
        
        Raster srcI = src[0].getRaster();			// intensity
        Raster srcH = src[1].getRaster();			// hue
        Raster srcS = src[2].getRaster();			// saturation
        
        WritableRaster wrR = biR.getRaster();	// red
        WritableRaster wrG = biG.getRaster();	// green
        WritableRaster wrB = biB.getRaster(); // blue
        
        updateBarra("Convertendo imagem, Aguarde!",30);
        System.out.println("Convertendo imagem...");
        
        for (int i = 0; i < hI; i++) {
            updateBarra("Convertendo imagem, Aguarde!", (int) (31 + (40*i)/hI));
            
            for (int j = 0; j < wI; j++) {
                // L os pixels de cada banda
                iI = srcI.getSample(j, i, 0);
                iH = srcH.getSample(j, i, 0);
                iS = srcS.getSample(j, i, 0);
                
                // Normaliza intensity e saturation, de [0, 255] para [0.0, 1.0]
                // hue pode ser qualquer nmero em [0.0, 1.0]
                intensity  = (float)iI/255.0f;
                hue        = (float)iH/255.0f;
                saturation = (float)iS/255.0f;
                
                // Converte HSB para RGB
                rgb = Color.HSBtoRGB(hue, saturation, intensity);
                iR = (rgb>>16)&0xFF;
                iG = (rgb>>8)&0xFF;
                iB = rgb&0xFF;
                
                // Grava pixels R, G, B
                wrR.setSample(j, i, 0, iR);
                wrG.setSample(j, i, 0, iG);
                wrB.setSample(j, i, 0, iB);
            }
        }
        out[0] = biR;
        out[1] = biG;
        out[2] = biB;
        
        return out;
    }
    
    // Salva imagens I, H, S no formato .PNG
    public File[] salvaImagem(BufferedImage[] dest, String nomeArquivo) {
        
        String search, prefixo, nome;
        File vFile[] = new File[3];
        
        // Grava imagem.
        search = ".";
        int i = nomeArquivo.indexOf(search);
        prefixo = nomeArquivo.substring(0, i);
        
        try {
            // imagem Hue Intensity
            updateBarra("Salvando as Imagens, Aguarde!",80);
            nome = "imagens\\" + prefixo + "_RGB_R.png";
            System.out.println("\nGravou arquivo -> " + nome + "\n");
            vFile[0] = new File(nome);
            ImageIO.write(dest[0], "png", vFile[0]);
            
            // imagem Intensity (Lightness)
            updateBarra("Salvando as Imagens, Aguarde!",85);
            nome = "imagens\\" + prefixo + "_RGB_G.png";
            System.out.println("\nGravou arquivo -> " + nome + "\n");
            vFile[1] = new File(nome);
            ImageIO.write(dest[1], "png", vFile[1]);
            
            // imagem Saturation
            updateBarra("Salvando as Imagens, Aguarde!",90);
            nome = "imagens\\" + prefixo + "_RGB_B.png";
            System.out.println("\nGravou arquivo -> " + nome + "\n");
            vFile[2] = new File(nome);
            ImageIO.write(dest[2], "png", vFile[2]);
        } catch (IOException e) {
            System.out.println("Problema gravando arquivo. ");
            System.exit(0);
        }
        return vFile;
    }
    
    // Abre arquivo de imagens
    private BufferedImage abreImagem(File aFile) {
        BufferedImage imagem = null;
        
        try {
            imagem = ImageIO.read(aFile);
        } catch(Exception e) {
            System.out.println("abreImagens: Problema lendo '" + aFile + "'.");
        }
        return imagem;
    }
    public boolean isSucesso() {
        return sucesso;
    }
}
